"""转中文有些问题，容易跳词"""


import re

from heartale.tts import HTS


class CoquiTTS(HTS):
    """获取待阅读文本的基础类
    """

    def __init__(self):
        """_summary_

        Args:
            key (str): 用于配置中区分使用本地什么服务
        """
        self.tts = None
        super().__init__("coqui")

    def set_conf(self, conf, py_libs=None):
        super().set_conf(conf, ["torch", "TTS"])
        from TTS.api import TTS  # pylint: disable=C0415
        self.tts = TTS(self.conf["model"]).to(self.conf["device"])

    async def download(self, text, file):

        text = re.sub(r'[“”]', '', text)
        text = re.sub(r'[…？！\n]', '。', text)
        if text[-1] != "。":
            text += "。"
        if self.tts is None:
            raise ValueError("初始化错误")
        self.tts.tts_to_file(text=text, file_path=file)



'常量'
from heartale.servers.legado import LegadoServer
from heartale.servers.txt import TxtServer
from heartale.tts.coqui import CoquiTTS
from heartale.tts.edge import EdgeTTS
from heartale.tts.fish import FishTTS
from heartale.tts.g_tts import GTTS
from heartale.tts.ms_azure import AzureTTS
from heartale.tts.paddle_speech import PaddleSpeechTTS


def get_servers():
    """获取所有服务

    Returns:
        dict: 所有服务
    """

    return [
        LegadoServer(),
        TxtServer(),
    ]


def get_ttses():
    """获取所有服务

    Returns:
        dict: 所有服务
    """

    return [
        EdgeTTS(),
        AzureTTS(),
        GTTS(),
        CoquiTTS(),
        PaddleSpeechTTS(),
        FishTTS()
    ]



"""fish speech"""

from heartale.tts import HTS


class FishTTS(HTS):
    """获取待阅读文本的基础类
    """

    def __init__(self):
        self.com = None
        self.data = {}
        super().__init__("fish")

    def set_conf(self, conf, py_libs=None):
        super().set_conf(conf, [])
        # pylint: disable=C0415
        from heartale.tts.fish.commons import ServeReferenceAudio
        from heartale.tts.fish.file import audio_to_bytes, read_ref_text

        references = []
        for a, t in self.conf.get("audio_text", {}).items():
            references.append(ServeReferenceAudio(
                audio=audio_to_bytes(a), text=read_ref_text(t)))

        self.data = {
            # "text": text,
            "references": references,
            "reference_id": None,
            "normalize": True,
            "format": "wav",
            "mp3_bitrate": 64,
            "opus_bitrate": -1000,
            "max_new_tokens": 0,
            "chunk_length": 200,
            "top_p": 0.7,
            "repetition_penalty": 1.2,
            "temperature": 0.7,
            "streaming": False,
            "use_memory_cache": "on-demand",
            "seed": None,
        }

    async def download(self, text, file):
        if text[-1] != "。":
            text += "。"
        # pylint: disable=C0415
        import aiohttp
        import ormsgpack

        # audio_text = {
        #     "/home/yuh/a13.mp3": "英老头一指点去，他的身躯突然炸开，四分五裂，元神被斩 ，死于非命。其他诸多神族心中悚然，急忙各自后退。"
        # }
        from heartale.tts.fish.commons import ServeTTSRequest
        self.data["text"] = text
        async with aiohttp.ClientSession() as session:
            async with session.post(
                self.conf["url_tts"],
                data=ormsgpack.packb(ServeTTSRequest(**self.data),
                                     option=ormsgpack.OPT_SERIALIZE_PYDANTIC),
                headers={
                    "authorization": "Bearer YOUR_API_KEY",
                    "content-type": "application/msgpack",
                },
            ) as response:
                if response.status == 200:
                    with open(file, "wb") as audio_file:
                        audio_file.write(await response.read())
                    print(f"Audio has been saved to '{file}'.")
                else:
                    print(f"Request failed with status code {response.status}")

                    error_data = await response.json()  # 异步读取 JSON
                    print(error_data)



"""文字转语音"""
import asyncio
import copy
import os

from heartale.tools import check_library_installed
from heartale.tools.config import get_config_tts_play


class HTS:
    """获取待阅读文本的基础类
    """

    def __init__(self, key: str):
        """_summary_

        Args:
            key (str): 用于配置中区分使用本地什么服务
        """

        self.key = key
        self.conf = {}

    def set_conf(self, conf, py_libs=None):
        """设置配置信息，并可以初始化一些服务，比如import第三方库，防止未使用的TTS导致内存占用过大

        Args:
            conf (dict): 配置信息
            py_libs (list, optional): 依赖的第三方库. Defaults to None.

        Raises:
            ImportError: _description_
        """
        self.conf = conf
        if py_libs is not None:
            for py_lib in py_libs:
                if not check_library_installed(py_lib):
                    raise ImportError(f"请安装{py_lib}库: pip install {py_lib}")

    async def download(self, text, file):
        """下载
        """
        print(text, file)


async def play_mp3(file_path, conf_all: dict):
    """子线程阅读

    Args:
        file_path (str): 音频文件路径
        conf_all (dict): 所有配置
    """

    codes = copy.copy(get_config_tts_play(conf_all)["code"])
    codes.append(file_path)

    process = await asyncio.create_subprocess_exec(*codes)
    await process.communicate()
    os.remove(file_path)



"""主文件"""
import asyncio
import os
import time
from datetime import datetime

from heartale.servers import Server
from heartale.tools.cache import get_cache_mp3, rm_cache_mp3
from heartale.tools.config import (get_config, get_config_server,
                                   get_config_tts_download)
from heartale.tools.constant import get_servers, get_ttses
from heartale.tools.count import save_read_time
from heartale.tts import HTS, play_mp3


def get_server(conf_all) -> Server:
    """基础类的扩展

    Args:
        conf_all (dict): 完整配置

    Returns:
        Server: _description_
    """
    key, conf_server = get_config_server(conf_all)

    for s in get_servers():
        if s.key == key:
            s.set_conf(conf_server)
            return s

    print(f"未知的服务 {key}")
    raise ValueError(f"未知的文本服务 {key}")


def get_tts(conf_all) -> HTS:
    """获取tts服务

    Args:
        conf_all (dict): _description_

    Returns:
        TTS: _description_
    """

    key, conf_tts = get_config_tts_download(conf_all)

    for tts in get_ttses():
        if tts.key == key:
            tts.set_conf(conf_tts)
            return tts

    raise ValueError(f"未知的TTS服务 {key}")


def print_test(i, chap, text, file):
    """_summary_

    Args:
        i (_type_): _description_
        chap (_type_): _description_
        text (_type_): _description_
        file (_type_): _description_
    """
    print(f"*** {i}/{chap} ***")
    print(file)
    if len(text) > 20:
        print(f"{text[:20]} ... {len(text)}")
        return

    print(text[:20])


async def play(chap=1000, play_min=100):
    """主函数，两个条件只要满足一个就停止

    Args:
        chap (int, optional): 默认听1000段落，自动停止. Defaults to 1000.
        play_min (int, optional): 默认播放100分钟，读完当前章节自动停止. Defaults to 100.
    """
    # https://github.com/gedoor/legado
    # 获取当前第一本书的信息
    st = time.time()

    # 本次阅读开始时间
    date_key = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    print(f"开始时间：{date_key}")
    # 每次阅读多久
    read_time_data = {
        "word": [],
        "time": []
    }

    rm_cache_mp3()

    conf = get_config()
    proxy_url = conf.get("proxy_url", "")

    if len(proxy_url) > 0:
        print(proxy_url)
        # 设置环境变量
        os.environ['http_proxy'] = proxy_url
        os.environ['https_proxy'] = proxy_url

    tts = get_tts(conf)
    server = get_server(conf)

    print(f"初始化完成：tts:{tts.key}, txt:{server.key}")
    print("下载中，请稍等……")

    # 下载和播放章节
    text = await server.initialize()
    mp3_file = f'{get_cache_mp3(f"{time.time()}")}.wav'
    await tts.download(text, mp3_file)

    # 默认听100章节，自动停止
    for _i in range(chap):
        # 默认播放100分钟，一段结束再停止
        if sum(read_time_data["time"]) > play_min * 60:
            print(f'阅读{(sum(read_time_data["time"]))/60} > {play_min}分钟')
            break

        read_time_data["time"].append(round(time.time() - st, 2))
        read_time_data["word"].append(len(text))
        save_read_time(date_key, read_time_data, server.book_name)
        st = time.time()

        # 并行播放和下载任务
        task_play = play_mp3(mp3_file, conf)

        text = await server.next()
        print_test(_i, chap, text, mp3_file)

        mp3_file = f'{get_cache_mp3(f"{time.time()}")}.wav'
        task_download = tts.download(text, mp3_file)

        await asyncio.gather(task_play, task_download)


def run_play():
    """运行主函数"""
    asyncio.run(play())



'fish-speech内置'
from typing import List, Literal, Optional

from pydantic import BaseModel


class ServeReferenceAudio(BaseModel):
    """_summary_

    Args:
        BaseModel (_type_): _description_
    """
    audio: bytes
    text: str

    def test(self):
        """_summary_
        """
        print("xx")

    def test2(self):
        """_summary_
        """
        print("xx")


class ServeTTSRequest(BaseModel):
    """_summary_

    Args:
        BaseModel (_type_): _description_
    """
    text: str
    # Annotated[int, conint(ge=100, le=300, strict=True)]
    chunk_length: int = 200
    # Audio format
    format: Literal["wav", "pcm", "mp3"] = "wav"
    mp3_bitrate: Literal[64, 128, 192] = 128
    # References audios for in-context learning
    references: List[ServeReferenceAudio] = []
    # Reference id
    # For example, if you want use https://fish.audio/m/7f92f8afb8ec43bf81429cc1c9199cb1/
    # Just pass 7f92f8afb8ec43bf81429cc1c9199cb1
    reference_id: str = None
    seed: int = None
    use_memory_cache: Literal["on-demand", "never"] = "never"
    # Normalize text for en & zh, this increase stability for numbers
    normalize: bool = True
    mp3_bitrate: Optional[int] = 64
    opus_bitrate: Optional[int] = -1000
    # Balance mode will reduce latency to 300ms, but may decrease stability
    latency: Literal["normal", "balanced"] = "normal"
    # not usually used below
    streaming: bool = False
    max_new_tokens: int = 1024
    # Annotated[float, Field(ge=0.1, le=1.0, strict=True)]
    top_p: float = 0.7
    # Annotated[float, Field(ge=0.9, le=2.0, strict=True)]
    repetition_penalty: float = 1.2
    # Annotated[float, Field(ge=0.1, le=1.0, strict=True)]
    temperature: float = 0.7

    def test(self):
        """_summary_
        """
        print("xx")

    def test2(self):
        """_summary_
        """
        print("xx")



"""文本转语音文件"""


from heartale.tts import HTS


class PaddleSpeechTTS(HTS):
    """获取待阅读文本的基础类
    """

    def __init__(self):
        self.tts = None
        super().__init__("paddlespeech")

    def set_conf(self, conf, py_libs=None):
        super().set_conf(conf, ["paddlespeech"])
        # pylint: disable=C0415
        from paddlespeech.cli.tts.infer import TTSExecutor
        self.tts = TTSExecutor()

    async def download(self, text, file):
        if self.tts is None:
            raise ValueError("初始化错误")
        self.tts(text=text, output=file)



"""文本转语音文件"""


from heartale.tools import get_proxy_url
from heartale.tts import HTS


class EdgeTTS(HTS):
    """获取待阅读文本的基础类
    """

    def __init__(self):
        self.com = None
        self.proxy = None
        self.err = None
        super().__init__("edge")

    def set_conf(self, conf, py_libs=None):
        super().set_conf(conf, ["edge_tts"])

        self.proxy = get_proxy_url()
        print(f"使用代理 {self.proxy}")

        import edge_tts  # pylint: disable=C0415
        self.com = edge_tts.Communicate
        self.err = edge_tts.exceptions

    async def download(self, text, file):
        if self.com is None or self.err is None:
            raise ValueError("初始化错误")
        try:
            await self.com(text, self.conf["voice"],
                           rate=self.conf["rate"],
                           proxy=self.proxy).save(file)
        except self.err.NoAudioReceived as e:
            print(f"下载失败：{text}\n{str(e)}")
            await self.download(f"文本空，{self.key} 跳过一次", file)
            return None



"""一些工具"""
import hashlib
import json
import os
import re
import time
from urllib.parse import quote


def split_text(text_all, dcp=0):
    """重新划分段落，防止太短

    Args:
        text_all (dict): 书json
        dcp (int, optional): 分段转音频，第几段. Defaults to 0.

    Returns:
        _type_: result（分割以后的文本数组）, p2s（每个字符在这个章节的索引位置）, n_last（记录的索引位置是第几个分割文本）
    """
    result = []
    p2s = [0]
    text = ""
    n_last = 0

    last = 0
    text_list = text_all.strip().split("\n")
    for i, line in enumerate(text_list):
        text += line + "\n"
        # 至少一段，如果一段没超过100个字，把下一段也连上，还不够100,继续
        # 如果这个章节最后，也算上
        if len(text) > 100 or i == len(text_list) - 1:

            if last <= dcp <= last + len(text):
                n_last = len(result)

            result.append(text)
            # 这个分割是第几个字符，方便保存进度
            p2s.append(last + len(text))

            last = last + len(text)
            text = ""

    if n_last > 0:
        n_last -= 1

    return result, p2s, n_last


def data2url(url):
    """将书籍信息URL编码

    Args:
        url (str): url

    Returns:
        str: 编码以后的图书信息url
    """
    return quote(url)


def cal_file_md5(file_path, chunk_size=8192):
    """计算文件md5

    Args:
        file_path (str): _description_
        chunk_size (int, optional): _description_. Defaults to 8192.

    Returns:
        str: _description_
    """
    st = time.time()
    # 创建一个MD5哈希对象
    md5_hash = hashlib.md5()

    # 按块读取文件并更新MD5对象
    with open(file_path, 'rb') as file:
        while chunk := file.read(chunk_size):
            md5_hash.update(chunk)

    # 获取MD5哈希值（以十六进制表示）
    print(f"计算文件md5耗时：{time.time() - st}")
    return md5_hash.hexdigest()


def save_data(file_path, data):
    """保存数据到文件

    Args:
        file_path (str): 文件路径
        data (_type_): 数据
    """

    with open(file_path, "w", encoding="utf-8") as f:
        json.dump(data, f, indent=4, ensure_ascii=False)


def get_data(file_path, data_defaul=None):
    """保存数据到文件

    Args:
        file_path (str): 文件路径
        data (_type_): 数据
    """

    if not os.path.exists(file_path) and data_defaul is not None:
        return data_defaul

    with open(file_path, "r", encoding="utf-8") as f:
        return json.load(f)


def check_library_installed(library_name):
    """某个库是否被安装

    Args:
        library_name (_type_): _description_

    Returns:
        _type_: _description_
    """
    try:
        __import__(library_name)
        return True
    except ImportError:
        return False


def detect_encoding(file_path):
    """
    Detect the encoding of a given file by attempting to read it with different encodings.

    Args:
        file_path (str): The path to the file whose encoding is to be detected.

    Returns:
        str: The detected encoding if successful, otherwise None.
    """

    encodings = ['utf-8', 'gbk', 'gb2312', 'big5']
    for encoding in encodings:
        try:
            with open(file_path, 'r', encoding=encoding) as file:
                file.read()
            return encoding
        except (UnicodeDecodeError, UnicodeError):
            continue

    return None  # 未能检测到编码


def parse_volumes_and_chapters(file_content, chap_n):
    """匹配 "第xx卷" 和 "第xx章"

    Args:
        file_content (str): _description_

    Returns:
        _type_: _description_
    """

    # 匹配 "第xx卷" 和 "第xx章"
    volume_pattern = r'^第([一二三四五六七八九十\d]+)卷\s*(.*)'  # 匹配卷号
    chapter_pattern = r'^第([一二三四五六七八九十百千\d]+)章\s*(.*)'  # 匹配章号

    current_volume = None
    chap_names = []
    p2s = []
    chap_content = ""

    # 按行解析文本
    # print(file_content)
    ss = file_content.split("\n")
    words = 0
    for line in ss:
        words += len(line + "\n")
        # 匹配卷号
        volume_match = re.search(volume_pattern, line)
        if volume_match:
            current_volume = volume_match.group()  # 获取当前卷
            continue  # 继续找章

        # 匹配章号
        chapter_match = re.search(chapter_pattern, line)
        if chapter_match:
            current_chapter = chapter_match.group()
            if current_volume:
                chap_names.append(f"{current_volume} {current_chapter}")
            else:
                chap_names.append(f"{current_chapter}")
            p2s.append(words)

        if len(p2s) == chap_n:
            chap_content += line + "\n"

    return chap_names, p2s, chap_content


def get_proxy_url():
    """获取代理地址

    Returns:
        str: 代理地址，如果没有设置代理则返回 None
    """
    http_p = os.getenv("http_proxy")
    if http_p:
        return http_p
    https_p = os.getenv("https_proxy")
    if https_p:
        return https_p

    return None



'申请github学生包，然后去微软官网申请，可以免费用一年'
# https://learn.microsoft.com/zh-cn/azure/ai-services/speech-service/language-support?tabs=tts

from heartale.tts import HTS


class AzureTTS(HTS):
    """文本转语音文件
    """

    def __init__(self):
        self.sc = None
        self.aoc = None
        self.ss = None
        self.rs = None
        super().__init__("azure")

    def set_conf(self, conf, py_libs=None):
        super().set_conf(conf, ["azure.cognitiveservices.speech"])

        if len(self.conf["key"]) == 0:
            raise ValueError("请在配置文件中填写Azure的key")
        # pylint: disable=C0415
        import azure.cognitiveservices.speech as speechsdk

        self.aoc = speechsdk.audio.AudioOutputConfig
        self.ss = speechsdk.SpeechSynthesizer
        self.rs = speechsdk.ResultReason.SynthesizingAudioCompleted

        self.sc = speechsdk.SpeechConfig(self.conf["key"], self.conf["region"])

    async def download(self, text, file):
        if self.ss is None or self.sc is None or self.aoc is None or self.rs is None:
            raise ValueError("初始化错误")
        ssml_text = f"""
        <speak version="1.0" xmlns="http://www.w3.org/2001/10/synthesis"
            xmlns:mstts="http://www.w3.org/2001/mstts" xml:lang="{self.conf['language']}">
            <voice name="{self.conf['voice']}">
                <prosody rate="{self.conf['rate']}">
                    {text}
                </prosody>
            </voice>
        </speak>
        """

        result = self.ss(self.sc, self.aoc(file))\
            .speak_ssml_async(ssml_text).get()
        if result.reason != self.rs:
            print(f"Speech synthesis failed: {result.reason}")
            raise ValueError(f"Speech synthesis failed: {result.reason}")

        return True



"""阅读本地txt文件"""
import os
import pathlib
from datetime import datetime

from heartale.servers import BookData, Server
from heartale.tools import (cal_file_md5, detect_encoding, get_data,
                            parse_volumes_and_chapters, save_data)
from heartale.tools.config import PATH_CONFIG_DIR

PATH_FILE = "path_file"
# 第几章
KEY_CHAP_N = "chap_n"
# 这一章第几个字
KEY_CHAP_TXT_POS = "chap_txt_pos"
KEY_ENCODING = "encoding"
LAST_READ_DATE = "last_read_date"


class TxtServer(Server):
    """阅读app相关的webapi"""

    def __init__(self):
        """初始化应用API

        Args:
            conf (dict): 配置 conf["legado"]
        """
        # 书籍位置
        self.path_file = ""
        # 编码格式
        self.ec = ""
        # 阅读进入文件位置
        self.path_read_p = ""
        self.bd = BookData()

        super().__init__("txt")

    async def initialize(self):
        """异步初始化"""
        if PATH_FILE not in self.conf:
            return "请设置待阅读的txt文件所在路径"

        self.path_file = self.conf[PATH_FILE]
        print(f"文件位置：{self.path_file}")
        os.makedirs(f"{PATH_CONFIG_DIR}/txts/", exist_ok=True)

        if not os.path.exists(self.path_file):
            self.bd.update_chap_txts("请检查设置的文件路径是否正确", 0)
            return "路径错误，文件不存在"

        file_ = pathlib.Path(self.path_file)
        self.book_name = file_.stem

        if file_.suffix != ".txt":
            self.bd.update_chap_txts("请检查设置的文件后缀名", 0)
            return f"此方式只支持txt文件，而不是{file_.suffix}"

        md5 = cal_file_md5(self.path_file)
        self.path_read_p = f"{PATH_CONFIG_DIR}/txts/{md5}.json"

        rp = self._get_read_progress()
        self._get_book_encoding(rp)
        chap_n = rp.get(KEY_CHAP_N, 0)
        chap_txt_pos = rp.get(KEY_CHAP_TXT_POS, 0)
        print(f"上次读取的位置：{chap_n}, {chap_txt_pos}")

        chap_names, chap_p2s, chap_content = parse_volumes_and_chapters(
            self._get_book_content(), chap_n)
        if chap_n == 0:
            chap_content = self._get_book_content()
        self.bd.set_chap_names(chap_names, chap_n, chap_p2s=chap_p2s)

        self.bd.update_chap_txts(chap_content, chap_txt_pos)
        return self.book_name + " " + self.bd.get_chap_name()

    async def next(self):
        """下一步

        Returns:
            str: 需要转音频的文本
        """
        print(f"当前位置：{self.bd.chap_txt_n}/{len(self.bd.chap_txts)}")

        if self.bd.is_chap_end():
            self.bd.chap_n += 1

            self.bd.update_chap_txts(
                self.bd.get_chap_content(self._get_book_content()))
            return self.bd.get_chap_name()

        txt = self.bd.chap_txts[self.bd.chap_txt_n]
        # 一些异常
        if len(self.bd.chap_txts) > 1:
            self._save_read_progress()
        self.bd.chap_txt_n += 1

        return txt

    def _get_read_progress(self):
        """读取阅读进度
        """
        data_default = {PATH_FILE: self.path_file,
                        KEY_CHAP_N: 0,
                        KEY_CHAP_TXT_POS: 0}
        data = get_data(self.path_read_p, data_default)
        if data[PATH_FILE] != self.path_file:
            print(f"文件位置不一致：{data[PATH_FILE]} -> {self.path_file}")

        return data

    def _save_read_progress(self):
        """异步保存阅读进度

        Args:
            book_data (dict): 书籍信息

        Raises:
            ValueError: 当进度保存出错时抛出异常
        """

        # 构建请求数据
        data = {
            PATH_FILE: self.path_file,
            KEY_CHAP_N: self.bd.chap_n,
            KEY_CHAP_TXT_POS: self.bd.get_chap_txt_pos(),
            KEY_ENCODING: self.ec,
            LAST_READ_DATE: datetime.now().isoformat(),
        }
        save_data(self.path_read_p, data)

    def _get_book_encoding(self, rp):
        self.ec = detect_encoding(self.path_file)
        if self.ec is None:
            self.ec = self.conf["encoding"]
            if len(self.ec) == 0:
                if KEY_ENCODING not in rp:
                    self.ec = "utf-8"
                self.ec = rp[KEY_ENCODING]

    def _get_book_content(self):
        file_content = ""
        try:
            with open(self.path_file, "r", encoding=self.ec) as f:
                file_content = f.read()
        except (UnicodeDecodeError, UnicodeError):
            with open(self.path_file, "r", encoding=self.ec, errors='ignore') as f:
                file_content = f.read()
            return f"{self.book_name}，txt文件使用{self.ec}解码可能出现错误，请修改配置中解码方法，或核对txt文件"
        return file_content



"""阅读app 相关的webapi"""
import datetime
import json
import time

import aiohttp

from heartale.servers import BookData, Server
from heartale.tools import data2url

# 常量定义
CHAP_POS = "durChapterPos"
CHAP_INDEX = "durChapterIndex"
CHAP_TITLE = "durChapterTitle"
CHAP_TXT_N = "durChapterTxtN"
TIMEOUT = aiohttp.ClientTimeout(total=10)


def get_base_url(conf):
    """_summary_

    Args:
        conf (_type_): _description_

    Returns:
        str: _description_
    """
    return f'http://{conf["ip"]}:{conf["port"]}'


def bu(book_data: dict):
    """_summary_

    Args:
        book_data (dict): _description_

    Returns:
        str: _description_
    """
    return f'url={data2url(book_data["bookUrl"])}'


class LegadoServer(Server):
    """阅读app相关的webapi"""

    def __init__(self):
        """初始化应用API

        Args:
            conf (dict): 配置 conf["legado"]
        """
        # 书籍信息
        self.book_data = {}
        self.bd = BookData()

        super().__init__("legado")

    async def initialize(self):

        self.book_data = await self._get_book_shelf(0)
        self.book_name = self.book_data["name"]

        self.bd.set_chap_names(await self._get_chapter_list(self.book_data),
                               self.book_data[CHAP_INDEX])
        self.bd.update_chap_txts(await self._get_book_txt(self.book_data),
                                 self.book_data[CHAP_POS])

        return self.book_name + " " + self.bd.get_chap_name()

    async def next(self):
        """下一步

        Returns:
            _type_: _description_
        """
        if self.bd.is_chap_end():
            self.bd.chap_n += 1
            s = self.bd.get_chap_name()

            self.book_data[CHAP_POS] = 0
            self.book_data[CHAP_INDEX] = self.bd.chap_n
            self.book_data[CHAP_TITLE] = s

            self.bd.update_chap_txts(await self._get_book_txt(self.book_data))

            return s

        txt = self.bd.chap_txts[self.bd.chap_txt_n]
        await self._save_book_progress(self.book_data)
        self.bd.chap_txt_n += 1

        return txt

    async def _get_book_shelf(self, book_n: int):
        """异步获取书架信息

        Args:
            book_n (int): 第几本书

        Returns:
            dict: 书籍信息
        """
        url = f"{get_base_url(self.conf)}/getBookshelf"

        async with aiohttp.ClientSession() as session:
            async with session.get(url, timeout=TIMEOUT) as response:
                resp_json = await response.json(content_type=None)

        return resp_json["data"][book_n]

    async def _get_chapter_list(self, book_data: dict):
        """异步获取书章节目录

        Args:
            book_data (dict): 书籍信息

        Returns:
            list: 章节目录，包含title等
        """
        url = f"{get_base_url(self.conf)}/getChapterList?{bu(book_data)}"

        async with aiohttp.ClientSession() as session:
            async with session.get(url, timeout=TIMEOUT) as response:
                resp_json = await response.json(content_type=None)

        return [d["title"] for d in resp_json["data"]]

    async def _get_book_txt(self, book_data: dict):
        """异步获取书某一章节的文本

        Args:
            book_data (dict): 书籍信息

        Returns:
            str: 某一章节的文字
        """
        url = f"{get_base_url(self.conf)}/getBookContent"
        params = f"{bu(book_data)}&index={book_data[CHAP_INDEX]}"

        async with aiohttp.ClientSession() as session:
            async with session.get(f"{url}?{params}", timeout=TIMEOUT) as response:
                resp_json = await response.json(content_type=None)

        return resp_json["data"]

    async def _save_book_progress(self, book_data: dict):
        """异步保存阅读进度

        Args:
            book_data (dict): 书籍信息

        Raises:
            ValueError: 当进度保存出错时抛出异常
        """
        # 获取当前时间戳（毫秒）
        dct = int(time.mktime(datetime.datetime.now().timetuple()) * 1000)

        # 构建请求数据
        data = {
            "name": self.book_name,
            "author": book_data["author"],
            CHAP_INDEX: self.bd.chap_n,
            CHAP_POS: self.bd.get_chap_txt_pos(),
            "durChapterTime": dct,
            CHAP_TITLE: self.bd.get_chap_name(),
        }

        json_data = json.dumps(data)
        headers = {'Content-Type': 'application/json'}

        async with aiohttp.ClientSession() as session:
            async with session.post(f"{get_base_url(self.conf)}/saveBookProgress",
                                    data=json_data,
                                    headers=headers,
                                    timeout=TIMEOUT) as response:
                resp_json = await response.json(content_type=None)

                if not resp_json["isSuccess"]:
                    raise ValueError(f'进度保存错误！\n{resp_json["errorMsg"]}')



"""配置"""
import os
import sys

from heartale.tools import get_data, save_data

if sys.platform.startswith('win'):
    PATH_CONFIG_DIR = os.path.join(os.getenv("APPDATA", ""), "heartale")
else:
    PATH_CONFIG_DIR = os.path.join(
        os.getenv("HOME", ""), ".config", "heartale")
os.makedirs(PATH_CONFIG_DIR, exist_ok=True)
print("config", PATH_CONFIG_DIR)
PATH_CONFIG = f'{PATH_CONFIG_DIR}/config.json'

CONFIG_DATA = None


DEFAULT_CONFIG = {
    "version": 1,
    "proxy_url": "",
    "server": {
        "key": "txt",
        "legado": {
            "ip": "192.168.1.6",
            "port": "1122"
        },
        "txt": {
            "path_file": "",
            "encoding": ""
        }
    },
    "tts": {
        "play": {
            "code": [
                "ffplay",
                "-nodisp",
                "-autoexit",
                "-loglevel",
                "quiet"
            ]
        },
        "download": {
            "key": "edge",
            "edge": {
                "voice": "zh-CN-XiaoxiaoNeural",
                "rate": "+30%"
            },
            "azure": {
                "key": "",
                "region": "japanwest",
                "language": "zh-CN",
                "voice": "zh-CN-XiaoxiaoNeural",
                "rate": "+30%"
            },
            "gtts": {
                "language": "zh-CN"
            },
            "coqui": {
                "model": "tts_models/zh-CN/baker/tacotron2-DDC-GST",
                "device": "cpu"
            },
            "paddlespeech": {},
            "fish": {
                "url_tts": "http://127.0.0.1:8080/v1/tts",
                "audio_text": {}
            }
        }
    }
}


def get_config(path=PATH_CONFIG):
    """获取配置

    Args:
        path (_type_, optional): _description_. Defaults to PATH_CONFIG.

    Returns:
        dict: 配置数据
    """

    if not os.path.exists(path):
        print("配置文件不存在，已创建默认配置文件")
        save_data(PATH_CONFIG, DEFAULT_CONFIG)
        return DEFAULT_CONFIG

    data = get_data(path)
    if "version" not in data or data["version"] != DEFAULT_CONFIG["version"]:
        save_data(PATH_CONFIG, DEFAULT_CONFIG)
        return DEFAULT_CONFIG

    return data


def get_config_server(conf_all):
    """获取服务配置

    Args:
        conf_all (dict): 配置数据

    Returns:
        dict: 服务配置
    """

    conf_servers = conf_all["server"]
    conf_server_key = conf_servers["key"]
    conf_server = conf_servers[conf_server_key]

    return conf_server_key, conf_server


def get_config_tts_play(conf_all):
    """获取tts配置

    Args:
        conf_all (dict): 配置数据

    Returns:
        dict: tts配置
    """

    return conf_all["tts"]["play"]


def get_config_tts_download(conf_all):
    """获取tts配置

    Args:
        conf_all (dict): 配置数据

    Returns:
        dict: tts配置
    """

    conf_ttss = conf_all["tts"]["download"]
    conf_tts_key = conf_ttss["key"]
    conf_tts = conf_ttss[conf_tts_key]
    return conf_tts_key, conf_tts



'''配置'''

import os
import shutil
import sys

if sys.platform.startswith('win'):
    ld = os.getenv('LOCALAPPDATA', "") or os.getenv('APPDATA', "")
    PATH_CACHE_DIR = os.path.join(ld, "heartale")
else:
    xdg_cache_home = os.getenv('XDG_CACHE_HOME')
    if xdg_cache_home is None:
        home = os.path.expanduser("~")
        xdg_cache_home = os.path.join(home, '.cache')
    PATH_CACHE_DIR = os.path.join(xdg_cache_home, "heartale")
os.makedirs(PATH_CACHE_DIR, exist_ok=True)


def get_cache_mp3(file):
    """_summary_

    Args:
        file (_type_): _description_

    Returns:
        _type_: _description_
    """
    return f"{PATH_CACHE_DIR}/mp3/{file}"


os.makedirs(get_cache_mp3(""), exist_ok=True)


def rm_cache_mp3():
    """清理缓存
    """
    mp3_path = get_cache_mp3("")
    if os.path.exists(mp3_path):
        shutil.rmtree(mp3_path)
    os.mkdir(mp3_path)



'统计阅读时间'
from datetime import datetime

from heartale.tools import get_data, save_data
from heartale.tools.config import PATH_CONFIG_DIR

PATH_READ_TIME = f"{PATH_CONFIG_DIR}/read_time.json"


def save_read_time(date_k, read_time_data, book_name):
    """保存阅读时间，每一天是一个key，每天可以多本书，每本书多次阅读，每次阅读多个时长间隔

    Args:
        date_k (str): 开始阅读时间
        date_v (list[int]): 每个循环阅读时间

    Returns:
        dict: {2024-10-20:{'2024-10-20 10:40:31': [1, 2, 3]}}
    """

    # 获取当前日期，只保留年月日
    today_key = str(datetime.now().date())

    data = get_data(PATH_READ_TIME, {})

    if today_key not in data:
        data[today_key] = {book_name: {date_k: read_time_data}}

    if book_name not in data[today_key]:
        data[today_key][book_name] = {date_k: read_time_data}

    # 时间列表一定替换，因为每次阅读date_k不一样，但是不同循环date_k一样
    data[today_key][book_name][date_k] = read_time_data

    save_data(PATH_READ_TIME, data)


def format_time(seconds):
    """计算天、小时、分钟和秒

    Args:
        seconds (float): _description_

    Returns:
        str: _description_
    """
    days, seconds = divmod(seconds, 86400)  # 86400秒 = 1天
    hours, seconds = divmod(seconds, 3600)  # 3600秒 = 1小时
    minutes, seconds = divmod(seconds, 60)  # 60秒 = 1分钟

    # 根据时间长度选择合适的显示格式
    if days > 0:
        return f"{days}天 {hours}小时 {minutes}分钟 {seconds:.2f}秒"
    if hours > 0:
        return f"{hours}小时 {minutes}分钟 {seconds:.2f}秒"
    if minutes > 0:
        return f"{minutes}分钟 {seconds:.2f}秒"

    return f"{seconds:.2f}秒"


def count_read_time():
    """显示详细信息
    """
    data = get_data(PATH_READ_TIME, {})
    book_names = {}
    for data_day, data_book in data.items():
        for book_name, data_time in data_book.items():
            if book_name not in book_names:
                book_names[book_name] = {
                    "read_time": 0,
                    "read_word": 0,
                    "days": {}
                }

            book_names[book_name]["days"][data_day] = {
                "read_time": 0,
                "read_word": 0
            }
            for _time_k, time_v in data_time.items():
                book_names[book_name]["read_time"] += sum(time_v["time"])
                book_names[book_name]["read_word"] += sum(time_v["word"])
                book_names[book_name]["days"][data_day]["read_time"] += sum(
                    time_v["time"])
                book_names[book_name]["days"][data_day]["read_word"] += sum(
                    time_v["word"])

    print(list(book_names.keys()))

    t = 0
    for book_name, data in book_names.items():
        t += data['read_time']
        s = f"\n***** {book_name} *****\n"
        s += f"总阅读时间：{format_time(data['read_time'])}"
        s += f"，总阅读字数：{data['read_word']}个"
        for day, d in data["days"].items():
            s += f"\n{day}，阅读时间：{format_time(d['read_time'])}"
            s += f"，阅读字数：{d['read_word']}个"
        print(s)
    print("\n******\n总阅读时间：", format_time(t))



'''获取文本并自动跳转的配置'''

from heartale.tools import split_text


class BookData():
    """_summary_
    """

    def __init__(self):
        # 章节目录
        self.chap_names = []
        # 章节目录所在位置
        self.chap_p2s = []
        # 这次是从第几个章节开始读的
        self.chap_n0 = 0
        # 现在是第几个章节
        self.chap_n = 0
        # 某章节的文本分割
        self.chap_txts = []
        # 某章节的文本分割所在位置
        self.chap_txt_p2s = [0]
        # 某章节的文本分割位置
        self.chap_txt_n = 0

    def set_chap_names(self, chap_names, chap_n, chap_p2s=None):
        """初始化数据

        Args:
            chap_names (list): 章节名
            chap_n (int): 上次读到哪个章节了
            chap_txt_pos (int, optional): 上次读到这个的哪个位置. Defaults to -1.
            chap_p2s (list, optional): 章节在全文中的分割位置. Defaults to None.
        """
        # 防止保存数据太多
        self.chap_n0 = chap_n
        self.chap_n = self.chap_n0

        self.chap_names = chap_names[self.chap_n:200+self.chap_n]
        if chap_p2s is not None:
            self.chap_p2s = chap_p2s[self.chap_n:200+self.chap_n]
        else:
            self.chap_p2s = None

    def get_chap_name(self):
        """获取章节名字"""
        return self.chap_names[self.chap_n-self.chap_n0]

    def update_chap_txts(self, chap_content, chap_txt_pos=0):
        """分割章节的文本

        Args:
            chap_content (str): 章节文本
            chap_txt_pos (int): 已经读到这个章节的什么位置
        """
        self.chap_txts, self.chap_txt_p2s, self.chap_txt_n = \
            split_text(chap_content, chap_txt_pos)

    def get_chap_content(self, book_content):
        """获取self.chap_n的章节内容

        Args:
            book_content (str): 书的完整内容

        Returns:
            str: self.chap_n章节的内容
        """
        if self.chap_p2s is None:
            raise ValueError("未设置章节位置")
        n = self.chap_n - self.chap_n0
        return book_content[self.chap_p2s[n]:self.chap_p2s[n+1]]

    def get_chap_txt_pos(self):
        """本章节的位置，需要保存

        Returns:
            _type_: _description_
        """
        return self.chap_txt_p2s[self.chap_txt_n]

    def is_chap_end(self):
        """这一章节是不是要结束了

        Returns:
            _type_: _description_
        """
        if self.chap_txt_n >= len(self.chap_txts):
            return True
        return False


class Server:
    """获取待阅读文本的基础类
    """

    def __init__(self, key: str):
        """_summary_

        Args:
            key (str): 用于配置中区分使用本地什么服务
        """

        self.key = key
        # 书名
        self.book_name = ""
        self.conf = {}

    def set_conf(self, conf):
        """设置配置信息"""
        self.conf = conf

    async def initialize(self):
        """异步初始化一些操作

        Returns:
            str: 比如书名等待阅读的文本
        """
        if not self.conf:
            print("请先设置配置信息")
        return "initialize"

    async def next(self):
        """接下来要阅读的文本，并保存本地阅读进度等信息
        """
        print("next")
        return "每次调用请自动刷新文本，并保存阅读信息"



"""不推荐，很机器人"""


from heartale.tts import HTS


class GTTS(HTS):
    """获取待阅读文本的基础类
    """

    def __init__(self):
        self.gtts = None

        super().__init__("gtts")

    def set_conf(self, conf, py_libs=None):
        super().set_conf(conf, ["gtts"])

        from gtts import gTTS  # pylint: disable=C0415
        self.gtts = gTTS

    async def download(self, text, file):
        if self.gtts is None:
            raise ValueError("初始化错误")
        self.gtts(text, lang=self.conf["language"]).save(file)
